#include "..\..\common\ImageFilter.h"
#include "SettingsForm.h"

using namespace FilterSpace;
using namespace System::Windows::Forms;
using namespace std;

#define DOUBLESHORTINT_TO_INT(a,b)     ((a) << 16 | ((b) & 0xffff))
#define FIRST_OF_DOUBLESHORTINT(a)     ((a) >> 16)
#define SECOND_OF_DOUBLESHORTINT(a)    ((a) & 0xffff)

int recognition_directions[2][8] = { { 1, 1, 0, -1, -1, -1, 0, 1 }, { 0, 1, 1, 1, 0, -1, -1, -1 } };




bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// WRITE YOUR FILTER CODE HERE -->
	int   block_count;

	int   pixel_index_check = f->pixel_index_check;
	int   pixel_index_new   = f->pixel_index_new;
	int   radius            = f->pixel_radius;
	int   surface_remove    = f->pixel_remove_surfaces;
	int   minimum_occurence = f->pixel_minimum_occurence;
	int   width             = f->width;
	int   height            = f->height;
	int   stride            = f->stridein;
	int   mstride           = (width + 7) >> 3;
	byte *din               = f->datain;
	byte *dout              = f->dataout;

	if (din != dout) memcpy(dout, din, stride * height);

	for (int y = radius; y < height - radius; y++)
	{
		for (int x = radius; x < width - radius; x++)
		{
			byte& color = dout[y*stride+x];
			if (color == pixel_index_check) // ha ref_color szn a pixel, nzzk meg a krnyezett
			{
				block_count = 0;
				for (int my = -radius; my <= radius; my++)
				for (int mx = -radius; mx <= radius; mx++)
				if (dout[(y + my) * stride + (x + mx)] == pixel_index_check) block_count++;
				if (block_count < minimum_occurence) color = pixel_index_new;
			}
			else if (color < 128 && color != pixel_index_check) // ha base_color szn, akkor legyen to_color
			{
				color = pixel_index_new;
			}
		}

		// FOR PERFORMANCE MEASUREMENT
		ImageFilterBase::UpdateFilterStatus(f, width, updater);
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	return true;
}




IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, stride = 0, height = 0, run_local = 1;

	if (!fs) return IntPtr(0);

	bool cancelled = false;
	xml_node^ xn = gcnew xml_node( fs );

	filtersettings* f = XmlToMyStruct( xn, true );

	if (f && !silent)
	{
		FilterTemplate::SettingsForm^ form = gcnew FilterTemplate::SettingsForm(f);
		form->ShowDialog();
		cancelled = form->FilterCancelled;
		delete form;
	}

	if (cancelled)
	{
		delete f;
		delete xn;
		return IntPtr(0);
	}

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* fb = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(fb);
	delete [] fb;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			xt->XmlGetChildValueByName("Stride", stride);
			xt->XmlGetChildValueByName("Height", height);
			if (stride * height == 0) return IntPtr(0);
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", outdata);
		}
	}
	xn->children.Add( xt );

	xn->XmlSetChildValueByName("PixelIndexCheck",       f->pixel_index_check);
	xn->XmlSetChildValueByName("PixelIndexNew",         f->pixel_index_new);
	xn->XmlSetChildValueByName("PixelRadius",           f->pixel_radius);
	xn->XmlSetChildValueByName("PixelMinimumOccurence", f->pixel_minimum_occurence);
	xn->XmlSetChildValueByName("PixelRemoveSurfaces",   f->pixel_remove_surfaces);
	xn->XmlSetChildValueByName("nCore!ExecutionSize",   1); // PIXELWISE == 1 ALWAYS FOR THIS FILTER

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}


filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	xml_node^ xntmp, ^xt2;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;

	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) return 0;

	if (xntmp->XmlGetChildByName(xt2, "ChannelsBitAllocation"))
	{
		String^ colMap;
		if (xt2->XmlGetChildValueByName("ColorMap", colMap))
		{
			unsigned char* data;
			int length;
			base64(StringTostring(colMap), data, length);
			memcpy(filset->palette256, data, MIN(1024,length));
		}
	} 

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	xn->XmlSetChildValueByName("PixelIndexCheck",       0, true);
	xn->XmlGetChildValueByName("PixelIndexCheck",       filset->pixel_index_check);
	xn->XmlSetChildValueByName("PixelIndexNew",         0, true);
	xn->XmlGetChildValueByName("PixelIndexNew",         filset->pixel_index_new);
	xn->XmlSetChildValueByName("PixelRadius",           1, true);
	xn->XmlGetChildValueByName("PixelRadius",           filset->pixel_radius);
	xn->XmlSetChildValueByName("PixelMinimumOccurence", 1, true);
	xn->XmlGetChildValueByName("PixelMinimumOccurence", filset->pixel_minimum_occurence);
	xn->XmlSetChildValueByName("PixelRemoveSurfaces",   1, true);
	xn->XmlGetChildValueByName("PixelRemoveSurfaces",   filset->pixel_remove_surfaces);

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) ) ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) ) ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;

	return filset;
}



bool ImageFilter::CanLocalize(){ return true; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName,  width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}


// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr)	= 0;
		*(filset->processing_size)	= filset->width * filset->height;
		b = (::FilterCore(filset, 0, 0, 0, 0, nullptr));
		delete filset;
		*(filset->processing_ptr)  = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}
